import java.util.Arrays;

class Curl {

    static final int HASH_LENGTH = 243;
    static final int STATE_LENGTH = HASH_LENGTH * 3;
    static final int NUMBER_OF_ROUNDS = 27;

    final byte[] state = new byte[STATE_LENGTH];

    void initialize(final byte[] length) { // "length" must contain the message length converted into HASH_LENGTH trits

        for (int i = 0; i < HASH_LENGTH; i++) {

            state[HASH_LENGTH * 0 + i] = 0;
            state[HASH_LENGTH * 1 + i] = length[i];
            state[HASH_LENGTH * 2 + i] = 0;
        }
    }

    void absorb(final byte[] trits, int offset, int length) {

        do {

            for (int i = 0; i < HASH_LENGTH && length-- > 0; i++) {

                state[i] = trits[offset++];
            }
            transform();

        } while (length > 0);
    }

    void squeeze(final byte[] trits, int offset, int length) {

        do {

            for (int i = 0; i < HASH_LENGTH && length-- > 0; i++) {

                trits[offset++] = state[i];
            }
            transform();

        } while (length > 0);
    }

    void transform() {

        for (int round = 0; round < NUMBER_OF_ROUNDS; round++) {

            for (int period = HASH_LENGTH; period > 0; period /= 3) {

                final byte[] scratchpad = new byte[STATE_LENGTH];

                for (int i = 0; i < STATE_LENGTH; i++) {

                    scratchpad[i] = state[i];
                }

                for (int i = 0; i < STATE_LENGTH; i++) {

                    final int indexA = i + (((i / period) + 0) % 3 - ((i / period) % 3)) * period;
                    final int indexB = i + (((i / period) + 1) % 3 - ((i / period) % 3)) * period;
                    final int indexC = i + (((i / period) + 2) % 3 - ((i / period) % 3)) * period;

                    state[i] = f(scratchpad[indexA], scratchpad[indexB], scratchpad[indexC]);
                }
            }
        }
    }

    static byte f(final byte a, final byte b, final byte c) {

        return (byte) ((
                (
                a*a*b*b*c + a*a*b*c*c - a*b*b*c*c
                + a*a*b*b - a*a*b*c + a*a*c*c + a*b*b*c
                - a*a*c + a*b*b - a*c*c + b*b*c + b*c*c
                - a*a - b*b + b*c - c*c
                - c
                + 1
                )
                + 27 + 1) % 3 - 1);
    }
}

public class Curl27Test {
    public static void curlHash(final byte[] hash, final byte[] msg)
    {
        byte[] length = new byte[Curl.HASH_LENGTH];
        int len = msg.length;
        for(int i = 0; i < Curl.HASH_LENGTH; i++)
        {
            byte l = (byte)(len % 3);
            if(l > 1) l = -1;
            len -= l;
            len /= 3;
            length[i] = l;
        }

        Curl c = new Curl();
        c.initialize(length);
        len = (msg.length + 242) / 243 * 243;
        c.absorb(Arrays.copyOf(msg, len), 0, len);
        c.squeeze(hash, 0, 243);
    }
    
    public static void main(String[] args) {
        byte[] msg = new byte[]
        {
            0, 1, -1, 0, 1, -1, 0, 1, -1, 
            0, 1, -1, 0, 1, -1, 0, 1, -1, 
            0, 1, -1, 0, 1, -1, 0, 1, -1, 
            0, 1, -1, 0, 1,
        };
        byte[] hash = new byte[243];
        curlHash(hash, msg);
        System.out.print("msg = ");
        System.out.println(Arrays.toString(msg));
        System.out.print("hash = ");
        System.out.println(Arrays.toString(hash));
    }
}
